@@ -3,33 +3,32 @@
 # Copyright (c) CERN, 2021.                 #
 # ######################################### #
 
-import numpy as np
 import logging
 from functools import partial
+from contextlib import contextmanager
 
-from .general import _pkg_root
-from .line_frozen import LineFrozen
-from .base_element import _handle_per_particle_blocks
-from .twiss import (twiss_from_tracker,
-                                 compute_one_turn_matrix_finite_differences,
-                                 find_closed_orbit, match_tracker
-                                )
-from .survey import survey_from_tracker
-from .internal_record import (new_io_buffer,
-                             start_internal_logging_for_elements_of_type,
-                             stop_internal_logging_for_elements_of_type)
-from .pipeline import PipelineStatus
-
+import numpy as np
 import xobjects as xo
 import xpart as xp
 
-from .beam_elements import Drift
-from .line import Line
-
 from . import linear_normal_form as lnf
+from .base_element import _handle_per_particle_blocks
+from .beam_elements import Drift
+from .general import _pkg_root
+from .internal_record import (new_io_buffer,
+                              start_internal_logging_for_elements_of_type,
+                              stop_internal_logging_for_elements_of_type)
+from .line import Line
+from .pipeline import PipelineStatus
+from .survey import survey_from_tracker
+from .tracker_data import TrackerData
+from .twiss import (compute_one_turn_matrix_finite_differences,
+                    find_closed_orbit, match_tracker, twiss_from_tracker)
+from .tapering import compensate_radiation_energy_loss
 
 logger = logging.getLogger(__name__)
 
+
 def _check_is_collective(ele):
     iscoll = not hasattr(ele, 'iscollective') or ele.iscollective
     return iscoll
@@ -56,7 +55,10 @@ class Tracker:
         io_buffer=None,
         compile=True,
         enable_pipeline_hold=False,
+        _element_ref_data=None,
     ):
+        self.config = TrackerConfig()
+        self.XTRACK_MULTIPOLE_NO_SYNRAD=True
 
         if sequence is not None:
             raise ValueError(
@@ -69,7 +71,14 @@ class Tracker:
                 self.iscollective = True
                 break
 
+        if not particles_monitor_class:
+            particles_monitor_class = self._get_default_monitor_class()
+
         if self.iscollective:
+            if _element_ref_data:
+                raise ValueError('The argument element_ref_data is not '
+                                 'supported in collective mode.')
+
             self._init_track_with_collective(
                 _context=_context,
                 _buffer=_buffer,
@@ -89,6 +98,7 @@ class Tracker:
                 compile=compile,
                 enable_pipeline_hold=enable_pipeline_hold)
         else:
+            self._element_ref_data = _element_ref_data
             self._init_track_no_collective(
                 _context=_context,
                 _buffer=_buffer,
@@ -171,8 +181,8 @@ class Tracker:
                 _element_index_in_part.append(ii_in_part)
                 ii_in_part += 1
             else:
-                if len(this_part.elements)>0:
-                    this_part.iscollective=False
+                if len(this_part.elements) > 0:
+                    this_part.iscollective = False
                     parts.append(this_part)
                     part_names.append(f'part_{i_part}_non_collective')
                     i_part += 1
@@ -183,8 +193,8 @@ class Tracker:
                 i_part += 1
                 this_part = Line(elements=[], element_names=[])
                 ii_in_part = 0
-        if len(this_part.elements)>0:
-            this_part.iscollective=False
+        if len(this_part.elements) > 0:
+            this_part.iscollective = False
             parts.append(this_part)
             part_names.append(f'part_{i_part}_non_collective')
 
@@ -192,8 +202,11 @@ class Tracker:
         noncollective_xelements = []
         for ii, pp in enumerate(parts):
             if not _check_is_collective(pp):
-                tempxtline = LineFrozen(_buffer=_buffer,
-                                   line=pp)
+                tempxtline = TrackerData(
+                    _buffer=_buffer,
+                    element_classes=element_classes,
+                    extra_element_classes=(particles_monitor_class._XoStruct,),
+                    line=pp)
                 pp.element_dict = dict(zip(
                     tempxtline.element_names, tempxtline.elements))
                 pp.element_names = tempxtline.element_names
@@ -223,6 +236,7 @@ class Tracker:
                 save_source_as=save_source_as,
                 io_buffer=self.io_buffer
                 )
+        supertracker.config = self.config
 
         # Build trackers for non collective parts
         for ii, pp in enumerate(parts):
@@ -238,6 +252,7 @@ class Tracker:
                                 local_particle_src=local_particle_src,
                                 skip_end_turn_actions=True,
                                 io_buffer=self.io_buffer)
+                parts[ii].config = self.config
 
         # Make a "marker" element to increase at_element
         self._zerodrift = Drift(_context=_buffer.context, length=0)
@@ -258,7 +273,6 @@ class Tracker:
         self._element_part = _element_part
         self._element_index_in_part = _element_index_in_part
 
-
     def _init_track_no_collective(
         self,
         _context=None,
@@ -279,87 +293,85 @@ class Tracker:
         compile=True,
         enable_pipeline_hold=False
     ):
+        if track_kernel == 'skip':
+            raise ValueError('Value `skip` for track kernel is deprecated for '
+                             'non-collective trackers, as all (re)compilation '
+                             'is done on demand.')
 
-        assert not(enable_pipeline_hold), (
-            "enable_pipeline_hold is not implemented in non collective mode")
+        if enable_pipeline_hold:
+            raise ValueError("`enable_pipeline_hold` is not implemented in "
+                             "non-collective mode")
         self._enable_pipeline_hold = False
 
         if particles_class is None:
             particles_class = xp.Particles
 
-        if particles_monitor_class is None:
-            import xtrack as xt  # I have to do it like this
-                                 # to avoid circular import #TODO to be solved
-            particles_monitor_class = xt.ParticlesMonitor
-
         if local_particle_src is None:
             local_particle_src = xp.gen_local_particle_api()
 
         self.global_xy_limit = global_xy_limit
         self.extra_headers = extra_headers
 
-        frozenline = LineFrozen(
-                    _context=_context, _buffer=_buffer, _offset=_offset,
-                    line=line)
+        tracker_data = TrackerData(
+            line=line,
+            element_classes=element_classes,
+            extra_element_classes=(particles_monitor_class._XoStruct,),
+            element_ref_data=self._element_ref_data,
+            _context=_context,
+            _buffer=_buffer,
+            _offset=_offset)
 
-        context = frozenline._buffer.context
+        context = tracker_data._buffer.context
 
         if io_buffer is None:
             io_buffer = new_io_buffer(_context=context)
         self.io_buffer = io_buffer
 
-        if track_kernel is None:
-            # Kernel relies on element_classes ordering
-            assert element_classes is None
+        if track_kernel is None and element_classes is not None:
+            raise ValueError('The kernel relies on `element_classes` ordering, '
+                             'so `element_classes` must be given if '
+                             '`track_kernel` is None.')
 
         if element_classes is None:
-            # Kernel relies on element_classes ordering
-            assert track_kernel=='skip' or track_kernel is None
-            element_classes = frozenline._ElementRefClass._reftypes + [
-                particles_monitor_class._XoStruct,
-            ]
+            if track_kernel is not None:
+                raise ValueError(
+                    'The kernel relies on `element_classes` ordering, so '
+                    '`track_kernel` must be given if `element_classes` is None.'
+                )
+            element_classes = tracker_data.element_classes
 
         line._freeze()
         self.line = line
         self.line.tracker = self
-        self._line_frozen = frozenline
-        ele_offsets = np.array(
-            [ee._offset for ee in frozenline.elements], dtype=np.int64)
-        ele_typeids = np.array(
-            [element_classes.index(ee._xobject.__class__)
-                for ee in frozenline.elements], dtype=np.int64)
-        ele_offsets_dev = context.nparray_to_context_array(ele_offsets)
-        ele_typeids_dev = context.nparray_to_context_array(ele_typeids)
+        self._tracker_data = tracker_data
 
         self.particles_class = particles_class
         self.particles_monitor_class = particles_monitor_class
-        self.ele_offsets_dev = ele_offsets_dev
-        self.ele_typeids_dev = ele_typeids_dev
-        self.num_elements = len(frozenline.elements)
+        self.num_elements = len(tracker_data.elements)
         self.global_xy_limit = global_xy_limit
         self.extra_headers = extra_headers
         self.skip_end_turn_actions = skip_end_turn_actions
         self.reset_s_at_end_turn = reset_s_at_end_turn
         self.local_particle_src = local_particle_src
         self.element_classes = element_classes
-        self._buffer = frozenline._buffer
+        self._buffer = tracker_data._buffer
 
-        if track_kernel == 'skip':
-            self.track_kernel = None
-        elif track_kernel is None:
-            self._build_kernel(save_source_as, compile=compile)
-        else:
-            self.track_kernel = track_kernel
+        if track_kernel is None:
+            track_kernel = {}
+        self.track_kernel = track_kernel
 
-        self.track=self._track_no_collective
+        self.track = self._track_no_collective
+
+        if compile:
+            _ = self._current_track_kernel # This triggers compilation
 
     def _invalidate(self):
         if self.iscollective:
-            self._invalidated_parts  = self._parts
+            self._invalidated_parts = self._parts
             self._parts = None
         else:
-            self._invalidated_line_frozen = self._line_frozen
-            self._line_frozen = None
+            self._invalidated_tracker_data = self._tracker_data
+            self._tracker_data = None
         self._is_invalidated = True
 
     def _check_invalidated(self):
@@ -370,7 +382,15 @@ class Tracker:
 
     def find_closed_orbit(self, particle_co_guess=None, particle_ref=None,
                           co_search_settings={}, delta_zeta=0, delta0=None,
-                          continue_on_closed_orbit_error=False):
+                          continue_on_closed_orbit_error=False,
+                          freeze_longitudinal=False):
+
+        if freeze_longitudinal:
+            kwargs = locals().copy()
+            kwargs.pop('self')
+            kwargs.pop('freeze_longitudinal')
+            with _freeze_longitudinal(self):
+                return self.find_closed_orbit(**kwargs)
 
         self._check_invalidated()
 
@@ -415,12 +435,15 @@ class Tracker:
         steps_r_matrix=None, co_search_settings=None, at_elements=None, at_s=None,
         values_at_element_exit=False,
         continue_on_closed_orbit_error=False,
+        freeze_longitudinal=False,
+        radiation_method='full',
         eneloss_and_damping=False,
-        ele_start=0, ele_stop=None, twiss_init=None,
+        ele_start=None, ele_stop=None, twiss_init=None,
         particle_on_co=None,
         matrix_responsiveness_tol=None,
         matrix_stability_tol=None,
-        symplectify=False
+        symplectify=False,
+        reverse=False,
         ):
 
         self._check_invalidated()
@@ -430,8 +453,10 @@ class Tracker:
 
         return twiss_from_tracker(self, **kwargs)
 
-    def survey(self,X0=0,Y0=0,Z0=0,theta0=0,phi0=0,psi0=0):
-        return survey_from_tracker(self,X0=X0,Y0=Y0,Z0=Z0,theta0=theta0,phi0=phi0,psi0=psi0)
+    def survey(self,X0=0,Y0=0,Z0=0,theta0=0,phi0=0,psi0=0, element0=0, reverse=False):
+        return survey_from_tracker(self, X0=X0, Y0=Y0, Z0=Z0, theta0=theta0,
+                                   phi0=phi0, psi0=psi0, element0=element0,
+                                   reverse=reverse)
 
     def match(self, vary, targets, **kwargs):
         return match_tracker(self, vary, targets, **kwargs)
@@ -449,6 +474,39 @@ class Tracker:
                  element_classes=(self.element_classes if not self.iscollective
                                     else self._supertracker.element_classes))
 
+    def configure_radiation(self, model=None, mode=None):
+
+        if mode is not None:
+            raise NameError('mode is deprecated, use model instead')
+
+        self._check_invalidated()
+
+        assert model in [None, 'mean', 'quantum']
+        if model == 'mean':
+            radiation_flag = 1
+        elif model == 'quantum':
+            radiation_flag = 2
+        else:
+            radiation_flag = 0
+
+        for kk, ee in self.line.element_dict.items():
+            if hasattr(ee, 'radiation_flag'):
+                ee.radiation_flag = radiation_flag
+
+        if radiation_flag == 2:
+            self.line._needs_rng = True
+
+        self.config.XTRACK_MULTIPOLE_NO_SYNRAD = False
+
+    def compensate_radiation_energy_loss(self, delta0=0, rtot_eneloss=1e-10,
+                                    max_iter=100, **kwargs):
+
+        all_kwargs = locals().copy()
+        all_kwargs.pop('self')
+        all_kwargs.pop('kwargs')
+        all_kwargs.update(kwargs)
+        compensate_radiation_energy_loss(self, **all_kwargs)
+
     def cycle(self, index_first_element=None, name_first_element=None,
               _buffer=None, _context=None):
 
@@ -478,6 +536,10 @@ class Tracker:
                 local_particle_src=self.local_particle_src,
             )
 
+    def build_particles(self, *args, **kwargs):
+        res = xp.build_particles(*args, tracker=self, **kwargs)
+        return res
+
     def get_backtracker(self, _context=None, _buffer=None,
                         global_xy_limit='from_tracker'):
 
@@ -518,10 +580,14 @@ class Tracker:
                 )
 
     @property
-    def particle_ref(self):
+    def particle_ref(self) -> xp.Particles:
         self._check_invalidated()
         return self.line.particle_ref
 
+    @particle_ref.setter
+    def particle_ref(self, value: xp.Particles):
+        self.line.particle_ref = value
+
     @property
     def vars(self):
         self._check_invalidated()
@@ -548,15 +614,9 @@ class Tracker:
     def _context(self):
         return self._buffer.context
 
-    def configure_radiation(self, mode=None):
-
-        self._check_invalidated()
-
-        self.line.configure_radiation(mode=mode)
-
     def _build_kernel(self, save_source_as, compile):
 
-        context = self._line_frozen._buffer.context
+        context = self._tracker_data._buffer.context
 
         kernels = {}
         headers = []
@@ -574,8 +634,7 @@ class Tracker:
             /*gpukern*/
             void track_line(
                 /*gpuglmem*/ int8_t* buffer,
-                /*gpuglmem*/ int64_t* ele_offsets,
-                /*gpuglmem*/ int64_t* ele_typeids,
+                             ElementRefData elem_ref_data,
                              ParticlesData particles,
                              int num_turns,
                              int ele_start,
@@ -617,16 +676,21 @@ class Tracker:
                     ParticlesMonitor_track_local_particle(tbt_monitor, &lpart);
                 }
 
-                for (int64_t ee=ele_start; ee<ele_start+num_ele_track; ee++){
+                int64_t elem_idx = ele_start;
+                for (; elem_idx < ele_start+num_ele_track; elem_idx++){
 
+                        #ifndef DISABLE_EBE_MONITOR
                         if (flag_monitor==2){
                             ParticlesMonitor_track_local_particle(tbt_monitor, &lpart);
                         }
+                        #endif
 
-                        /*gpuglmem*/ int8_t* el = buffer + ele_offsets[ee];
-                        int64_t ee_type = ele_typeids[ee];
+                        // Get the pointer to and the type id of the `elem_idx`th
+                        // element in `element_ref_data.elements`:
+                        void* el = ElementRefData_member_elements(elem_ref_data, elem_idx);
+                        int64_t elem_type = ElementRefData_typeid_elements(elem_ref_data, elem_idx);
 
-                        switch(ee_type){
+                        switch(elem_type){
         """
         )
 
@@ -655,11 +719,16 @@ class Tracker:
         src_lines.append(
             """
                         } //switch
+
+                    // Setting the below flag will break particle losses
+                    #ifndef DANGER_SKIP_ACTIVE_CHECK_AND_SWAPS
                     isactive = check_is_active(&lpart);
                     if (!isactive){
                         break;
                     }
                     increment_at_element(&lpart);
+                    #endif
+
                 } // for elements
                 if (flag_monitor==2){
                     // End of turn (element-by-element mode)
@@ -685,8 +754,7 @@ class Tracker:
             "track_line": xo.Kernel(
                 args=[
                     xo.Arg(xo.Int8, pointer=True, name="buffer"),
-                    xo.Arg(xo.Int64, pointer=True, name="ele_offsets"),
-                    xo.Arg(xo.Int64, pointer=True, name="ele_typeids"),
+                    xo.Arg(self._tracker_data._element_ref_data.__class__, name="tracker_data"),
                     xo.Arg(self.particles_class._XoStruct, name="particles"),
                     xo.Arg(xo.Int32, name="num_turns"),
                     xo.Arg(xo.Int32, name="ele_start"),
@@ -713,7 +781,7 @@ class Tracker:
         context.add_kernels(
             [source_track],
             kernels,
-            extra_headers=headers,
+            extra_headers=self._config_to_headers() + headers,
             extra_classes=self.element_classes,
             apply_to_source=[
                 partial(_handle_per_particle_blocks,
@@ -723,7 +791,7 @@ class Tracker:
             compile=compile
         )
 
-        self.track_kernel = context.kernels.track_line
+        self._current_track_kernel = context.kernels.track_line
 
     def _prepare_collective_track_session(self, particles, ele_start, ele_stop,
                                        num_elements, num_turns, turn_by_turn_monitor):
@@ -732,10 +800,10 @@ class Tracker:
         if particles.start_tracking_at_element >= 0:
             if ele_start != 0:
                 raise ValueError("The argument ele_start is used, but particles.start_tracking_at_element is set as well. "
-                                + "Please use only one of those methods.")
+                                 "Please use only one of those methods.")
             ele_start = particles.start_tracking_at_element
             particles.start_tracking_at_element = -1
-        if isinstance(ele_start,str):
+        if isinstance(ele_start, str):
             ele_start = self.line.element_names.index(ele_start)
 
         # ele_start can only have values of existing element id's,
@@ -878,6 +946,22 @@ class Tracker:
     def resume(self, session):
         return self._track_with_collective(particles=None, _session_to_resume=session)
 
+    def freeze_vars(self, variable_names):
+        for name in variable_names:
+            self.config[f'FREEZE_VAR_{name}'] = True
+
+    def unfreeze_vars(self, variable_names):
+        for name in variable_names:
+            self.config[f'FREEZE_VAR_{name}'] = False
+
+    def freeze_longitudinal(self, state=True):
+        assert state in (True, False)
+        assert self.iscollective is False, ('Cannot freeze longitudinal '
+                        'variables in collective mode (not yet implemented)')
+        if state:
+            self.freeze_vars(xp.particles.part_energy_varnames() + ['zeta'])
+        else:
+            self.unfreeze_vars(xp.particles.part_energy_varnames() + ['zeta'])
 
     def _track_with_collective(
         self,
@@ -887,8 +971,12 @@ class Tracker:
         num_elements=None, # defaults to full lattice
         num_turns=None,    # defaults to 1
         turn_by_turn_monitor=None,
+        freeze_longitudinal=False,
         _session_to_resume=None
     ):
+        if freeze_longitudinal:
+            raise NotImplementedError('freeze_longitudinal not implemented yet'
+                                      ' for collective tracking')
 
         self._check_invalidated()
 
@@ -1036,7 +1124,6 @@ class Tracker:
 
         self.record_last_track = monitor
 
-
     def _track_no_collective(
         self,
         particles,
@@ -1044,9 +1131,18 @@ class Tracker:
         ele_stop=None,     # defaults to full lattice
         num_elements=None, # defaults to full lattice
         num_turns=None,    # defaults to 1
-        turn_by_turn_monitor=None
+        turn_by_turn_monitor=None,
+        freeze_longitudinal=False
     ):
 
+        if freeze_longitudinal:
+            kwargs = locals().copy()
+            kwargs.pop('self')
+            kwargs.pop('freeze_longitudinal')
+
+            with _freeze_longitudinal(self):
+                return self._track_no_collective(**kwargs)
+
         self._check_invalidated()
 
         if isinstance(self._buffer.context, xo.ContextCpu):
@@ -1060,10 +1156,10 @@ class Tracker:
         if particles.start_tracking_at_element >= 0:
             if ele_start != 0:
                 raise ValueError("The argument ele_start is used, but particles.start_tracking_at_element is set as well. "
-                                + "Please use only one of those methods.")
+                                 "Please use only one of those methods.")
             ele_start = particles.start_tracking_at_element
             particles.start_tracking_at_element = -1
-        if isinstance(ele_start,str):
+        if isinstance(ele_start, str):
             ele_start = self.line.element_names.index(ele_start)
 
         assert ele_start >= 0
@@ -1156,13 +1252,12 @@ class Tracker:
         if self.line._needs_rng and not particles._has_valid_rng_state():
             particles._init_random_number_generator()
 
-        self.track_kernel.description.n_threads = particles._capacity
+        self._current_track_kernel.description.n_threads = particles._capacity
 
         # First turn
-        self.track_kernel(
-            buffer=self._line_frozen._buffer.buffer,
-            ele_offsets=self.ele_offsets_dev,
-            ele_typeids=self.ele_typeids_dev,
+        self._current_track_kernel(
+            buffer=self._tracker_data._buffer.buffer,
+            tracker_data=self._tracker_data._element_ref_data,
             particles=particles._xobject,
             num_turns=1,
             ele_start=ele_start,
@@ -1177,10 +1272,9 @@ class Tracker:
 
         # Middle turns
         if num_middle_turns > 0:
-            self.track_kernel(
-                buffer=self._line_frozen._buffer.buffer,
-                ele_offsets=self.ele_offsets_dev,
-                ele_typeids=self.ele_typeids_dev,
+            self._current_track_kernel(
+                buffer=self._tracker_data._buffer.buffer,
+                tracker_data=self._tracker_data._element_ref_data,
                 particles=particles._xobject,
                 num_turns=num_middle_turns,
                 ele_start=0, # always full turn
@@ -1195,10 +1289,9 @@ class Tracker:
 
         # Last turn, only if incomplete
         if num_elements_last_turn > 0:
-            self.track_kernel(
-                buffer=self._line_frozen._buffer.buffer,
-                ele_offsets=self.ele_offsets_dev,
-                ele_typeids=self.ele_typeids_dev,
+            self._current_track_kernel(
+                buffer=self._tracker_data._buffer.buffer,
+                tracker_data=self._tracker_data._element_ref_data,
                 particles=particles._xobject,
                 num_turns=1,
                 ele_start=0,
@@ -1213,6 +1306,13 @@ class Tracker:
 
         self.record_last_track = monitor
 
+    @staticmethod
+    def _get_default_monitor_class():
+        import xtrack as xt  # I have to do it like this
+
+        # to avoid circular import #TODO to be solved
+        return xt.ParticlesMonitor
+
     def _get_monitor(self, particles, turn_by_turn_monitor, num_turns):
 
         if turn_by_turn_monitor is None or turn_by_turn_monitor is False:
@@ -1255,3 +1355,130 @@ class Tracker:
     def stop_internal_logging_for_elements_of_type(self, element_type):
         self._check_invalidated()
         stop_internal_logging_for_elements_of_type(self, element_type)
+
+    def to_binary_file(self, path):
+        try:
+            tracker_data = self._tracker_data
+        except AttributeError:
+            raise TypeError("Only non-collective trackers can be binary serialized.")
+
+        # Serialise the tracker_data (line)
+        if not isinstance(tracker_data._context, xo.ContextCpu):
+            buffer = xo.ContextCpu().new_buffer(0)
+        else:
+            buffer = None
+
+        buffer, header_offset = tracker_data.to_binary(buffer)
+
+        # Serialise the knobs
+        var_management = {}
+        if tracker_data.line._var_management:
+            var_management = tracker_data.line._var_management_to_dict()
+
+        # Serialise the reference particle
+        particle_ref = None
+        if self.particle_ref:
+            particle_ref = self.particle_ref.to_dict()
+
+        with open(path, 'wb') as f:
+            np.save(f, header_offset)
+            np.save(f, buffer.buffer)
+            np.save(f, var_management, allow_pickle=True)
+            np.save(f, particle_ref, allow_pickle=True)
+
+    @classmethod
+    def from_binary_file(cls, path, particles_monitor_class=None, **kwargs) -> 'Tracker':
+        if not particles_monitor_class:
+            particles_monitor_class = cls._get_default_monitor_class()
+
+        with open(path, 'rb') as f:
+            header_offset = np.load(f)
+            np_buffer = np.load(f)
+            var_management_dict = np.load(f, allow_pickle=True).item()
+            particle_ref = np.load(f, allow_pickle=True).item()
+
+        xbuffer = xo.ContextCpu().new_buffer(np_buffer.nbytes)
+        # make sure that if we carry on using the buffer we
+        # don't overwrite things, by marking everything as used
+        xbuffer.allocate(np_buffer.nbytes)
+        xbuffer.buffer = np_buffer
+        tracker_data = TrackerData.from_binary(
+            xbuffer,
+            header_offset,
+            extra_element_classes=(particles_monitor_class,),
+        )
+        if var_management_dict:
+            tracker_data.line._init_var_management(var_management_dict)
+
+        tracker = Tracker(
+            line=tracker_data.line,
+            _element_ref_data=tracker_data._element_ref_data,
+            **kwargs,
+        )
+
+        if particle_ref is not None:
+            tracker.particle_ref = xp.Particles.from_dict(particle_ref)
+
+        return tracker
+
+    def _hashable_config(self):
+        items = ((k, v) for k, v in self.config.items() if v is not False)
+        return tuple(sorted(items))
+
+    def _config_to_headers(self):
+        headers = []
+        for k, v in self.config.items():
+            if not isinstance(v, bool):
+                headers.append(f'#define {k} {v}')
+            elif v is True:
+                headers.append(f'#define {k}')
+            else:
+                headers.append(f'#undef {k}')
+        return headers
+
+    @property
+    def _current_track_kernel(self):
+        try:
+            return self.track_kernel[self._hashable_config()]
+        except KeyError:
+            self._build_kernel(save_source_as=None, compile=True)
+            return self._current_track_kernel
+
+    @_current_track_kernel.setter
+    def _current_track_kernel(self, value):
+        self.track_kernel[self._hashable_config()] = value
+
+@contextmanager
+def _preserve_config(tracker):
+    config = TrackerConfig()
+    config.update(tracker.config)
+    try:
+        yield
+    finally:
+        tracker.config = config
+
+@contextmanager
+def freeze_longitudinal(tracker):
+    """Context manager to freeze longitudinal motion in a tracker."""
+    config = TrackerConfig()
+    config.update(tracker.config)
+    tracker.freeze_longitudinal(True)
+    try:
+        yield None
+    finally:
+        tracker.config = config
+_freeze_longitudinal = freeze_longitudinal # to avoid name clash with function argument
+
+class TrackerConfig(dict):
+    def __init__(self, *args, **kwargs):
+        super(TrackerConfig, self).__init__(*args, **kwargs)
+        object.__setattr__(self, '__dict__', self)
+
+    def __setitem__(self, idx, val):
+        if val is False and idx in self:
+            del(self[idx]) # We don't want to store flags that are False
+        else:
+            super(TrackerConfig, self).__setitem__(idx, val)
+
+    def __setattr__(self, idx, val):
+        self[idx] = val
